import os
import os.path

MODULE_DIR = os.path.split(os.path.abspath(__file__))[0]
PDB = os.path.join(MODULE_DIR, "md_1u19.pdb")
GRO = os.path.join(MODULE_DIR, "md_1u19.gro")
XTC = os.path.join(MODULE_DIR, "md_1u19.xtc")
TRR = os.path.join(MODULE_DIR, "md_1u19.trr")
ASE_Traj = os.path.join(MODULE_DIR, "md_1u19.traj")
ALA3 = os.path.join(MODULE_DIR, "ala3.pdb")



import json
import os

import nglview as nv


def test_make_sure_versions_matched():
    nglview_dir = os.path.join(os.path.dirname(__file__), '..')

    # frontend and backend
    nglview_js_widgets = os.path.join(nglview_dir, 'static', 'index.js')
    with open(nglview_js_widgets) as fh:
        for line in fh:
            if '"name":"nglview-js-widgets"' in line:
                break
    line = line.replace('module.exports =', '')
    assert json.loads(line)['version'] == nv.widget.__frontend_version__



from ipywidgets import DOMWidget
from traitlets import Bool, List


class BaseWidget(DOMWidget):
    _msg_q = []
    _msg_ar = List().tag(sync=True)
    _ready = Bool(False).tag(sync=True)

    def _js(self, code):
        self._call("executeCode", code)

    def _call(self, method_name, *args, **kwargs):
        msg = {"type": "callMethod",
               "methodName": method_name,
                "args": args, "kwargs": kwargs}
        if not self._ready:
            # fire later
            self._msg_q.append(msg)
        else:
            self.send(msg)
        msg_ar = self._msg_ar[:]
        msg_ar.append(msg)
        self._msg_ar = msg_ar # trigger sync



from io import StringIO

from . import datafiles
from .adaptor import (ASEStructure, ASETrajectory, BiopythonStructure,
                      FileStructure, HTMDTrajectory, IODataStructure,
                      IOTBXStructure, MDAnalysisTrajectory, MDTrajTrajectory,
                      OpenbabelStructure, ParmEdTrajectory, PdbIdStructure,
                      ProdyStructure, ProdyTrajectory, PyTrajTrajectory,
                      QCelementalStructure, RosettaStructure,
                      SchrodingerStructure, SchrodingerTrajectory,
                      TextStructure)
from .widget import NGLWidget

__all__ = [
    'demo',
    'show_pdbid',
    'show_url',
    'show_text',
    'show_ase',
    'show_pymatgen',
    'show_iotbx',
    'show_iodata',
    'show_qcelemental',
    'show_openbabel',
    'show_psi4',
    'show_rosetta',
    'show_asetraj',
    'show_simpletraj',
    'show_prody',
    'show_mdtraj',
    'show_pytraj',
    'show_mdanalysis',
    'show_parmed',
    'show_rdkit',
    'show_structure_file',
    'show_file',
    'show_htmd',
    'show_schrodinger',
    'show_biopython',
]


def show_pdbid(pdbid, **kwargs):
    '''Show PDB entry.

    Examples
    --------
    >>> import nglview as nv
    >>> w = nv.show_pdbid("3pqr")
    >>> w # doctest: +SKIP
    '''
    structure = PdbIdStructure(pdbid)
    return NGLWidget(structure, **kwargs)


def show_url(url, **kwargs):
    kwargs2 = {k: v for k, v in kwargs.items()}
    view = NGLWidget()
    view.add_component(url, **kwargs2)
    return view


def show_text(text, **kwargs):
    """for development
    """
    structure = TextStructure(text)
    return NGLWidget(structure, **kwargs)


def show_ase(ase_atoms, **kwargs):
    """

    Examples
    --------
    >>> import nglview as nv
    >>> from ase import Atom, Atoms
    >>> dimer = Atoms([Atom('X', (0, 0, 0)),
    ...                Atom('X', (0, 0, 1))])
    >>> dimer.set_positions([(1, 2, 3), (4, 5, 6.2)])
    >>> w = nv.show_ase(dimer)
    >>> w # doctest: +SKIP
    """
    structure = ASEStructure(ase_atoms)
    return NGLWidget(structure, **kwargs)


def show_iodata(obj, **kwargs):
    """Show iodata's IOData (require `ase` package).

    Examples
    --------
    >>> import nglview as nv
    >>> from iodata import IOData # doctest: +SKIP
    ... obj = IOData.from_file('what.xyz')
    ... view = nv.show_iodata(obj)
    ... view
    """
    return NGLWidget(IODataStructure(obj), **kwargs)


def show_qcelemental(obj, **kwargs):
    """Show QCelemental's Molecule (require `ase` package).

    Examples
    --------
    >>> import nglview as nv
    >>> import qcelemental as qcel # doctest: +SKIP
    ... mol = qcel.models.Molecule.from_data("He 0 0 0")
    ... view = nv.show_qcelemental(mol)
    ... view
    """
    return NGLWidget(QCelementalStructure(obj), **kwargs)


def show_psi4(obj, **kwargs):
    """Show psi4's Molecule (require `ase` package)

    Examples
    --------
    >>> import nglview as nv
    >>> import psi4 # doctest: +SKIP
    ... mol = psi4.geometry('xyz content here')
    ... view = nv.show_psi4(mol)
    ... view
    """
    return NGLWidget(QCelementalStructure(obj), **kwargs)


def show_openbabel(obj, **kwargs):
    """Show openbabel's Omol

    >>> import nglview
    >>> mol = openbabel.OBMol() # doctest: +SKIP
    ... obConversion = openbabel.OBConversion() 
    ... obConversion.SetInFormat('xyz')
    ... obConversion.ReadFile(mol, 'what.xyz')
    ... nglview.show_openbabel(mol)
    """
    return NGLWidget(OpenbabelStructure(obj), **kwargs)


def show_pymatgen(struct, **kwargs):
    """Require `ase` package.

    Examples
    --------
    >>> import nglview as nv # doctest: +SKIP
    ... import pymatgen as mg
    ... lattice = mg.Lattice.cubic(4.2)
    ... structure = mg.Structure(lattice, ["Cs", "Cl"],
                          [[0, 0, 0], [0.5, 0.5, 0.5]])
    ... view = nv.show_pymatgen(structure)
    ... view
    """
    from pymatgen.io.ase import AseAtomsAdaptor
    return show_ase(AseAtomsAdaptor().get_atoms(struct))


def show_iotbx(mol, **kwargs):
    """

    Examples
    --------
    >>> import iotbx.pdb
    ... x = iotbx.pdb.hierarchy.input('file.pdb')
    ... mol = x.construct_hierarchy()
    ... view = nglview.show_iotbx(hi)
    ... view # doctest: +SKIP
    """
    structure = IOTBXStructure(mol)
    return NGLWidget(structure, **kwargs)


def show_rosetta(pose, **kwargs):
    """

    Examples
    --------
    >>> from pyrosetta import pose_from_sequence, init
    ... init()
    ... pose = pose_from_sequence('AAAAAA')
    ... view = nglview.show_rosetta(pose)
    ... view # doctest: +SKIP
    """
    structure = RosettaStructure(pose)
    return NGLWidget(structure, **kwargs)


def show_asetraj(ase_traj, **kwargs):
    '''Show ase trajectory and structure file.

    Examples
    --------
    >>> import nglview as nv
    >>> from ase.io.trajectory import Trajectory
    >>> traj = Trajectory(nv.datafiles.ASE_Traj)
    >>> view = nv.show_asetraj(traj)
    >>> view.add_spacefill()
    >>> view # doctest: +SKIP
    '''
    trajectory = ASETrajectory(ase_traj)
    return NGLWidget(trajectory, **kwargs)


def show_structure_file(path, **kwargs):
    '''Show structure file. Allowed are text-based structure
    file formats that are by supported by NGL, including pdb,
    gro, mol2, sdf.

    Examples
    --------
    >>> import nglview as nv
    >>> w = nv.show_structure_file(nv.datafiles.GRO)
    >>> w # doctest: +SKIP
    '''
    structure = FileStructure(path)
    return NGLWidget(structure, **kwargs)


def _show_schrodinger_file(path, **kwargs):
    from schrodinger.structure import StructureReader

    view = NGLWidget()
    cts = list(StructureReader(path))
    for ct in cts:
        view.add_component(SchrodingerStructure(ct), **kwargs)
    return view


def show_file(path, **kwargs):
    '''Show any supported file format (e.g: .gro, .dx, ...)

    Examples
    --------
    >>> import nglview as nv
    ... w = nv.show_file('my.dx')
    ... w # doctest: +SKIP
    '''
    if isinstance(path, str) and (path.endswith(
        ('.mae', '.mae.gz', '.meagz', '.cms', '.cms.gz', '.cms.gz')) or
                                  (kwargs.get('format', None) == 'maestro')):
        view = _show_schrodinger_file(path, **kwargs)
    else:
        view = NGLWidget()
        view.add_component(path, **kwargs)
    return view


def show_simpletraj(traj, **kwargs):
    '''Show simpletraj trajectory and structure file.

    Examples
    --------
    >>> import nglview as nv
    >>> traj = nv.SimpletrajTrajectory(nv.datafiles.XTC, nv.datafiles.GRO)
    >>> view = nv.show_simpletraj(traj)
    >>> view # doctest: +SKIP
    '''
    return NGLWidget(traj, **kwargs)


def show_prody(obj, **kwargs):
    """

    Examples
    --------
    >>> import nglview as nv # doctest: +SKIP
    ... import prody
    ... structure = prody.parsePDB('what.pdb')
    ... ensemble = prody.parseDCD('what.dcd')
    ... ensemble.setAtoms(structure)
    ... nv.show_prody(ensemble)
    ... # nv.show_prody(structure)
    """
    import prody
    if isinstance(obj, prody.Ensemble):
        view_obj = ProdyTrajectory(obj)
    else:
        view_obj = ProdyStructure(obj)
    return NGLWidget(view_obj, **kwargs)


def show_mdtraj(mdtraj_trajectory, **kwargs):
    '''Show mdtraj trajectory.

    Examples
    --------
    >>> import nglview as nv # doctest: +SKIP
    ... import mdtraj as md
    ... t = md.load(nv.datafiles.XTC, top=nv.datafiles.GRO)
    ... w = nv.show_mdtraj(t)
    ... w
    '''
    structure_trajectory = MDTrajTrajectory(mdtraj_trajectory)
    return NGLWidget(structure_trajectory, **kwargs)


def show_pytraj(pytraj_trajectory, **kwargs):
    '''Show pytraj trajectory.

    Examples
    --------
    >>> import nglview as nv
    >>> import pytraj as pt
    >>> t = pt.load(nv.datafiles.TRR, nv.datafiles.PDB)
    >>> w = nv.show_pytraj(t)
    >>> w # doctest: +SKIP
    '''
    trajlist = pytraj_trajectory if isinstance(pytraj_trajectory,
                                               (list, tuple)) else [
                                                   pytraj_trajectory,
                                               ]

    trajlist = [PyTrajTrajectory(traj) for traj in trajlist]
    return NGLWidget(trajlist, **kwargs)


def show_parmed(parmed_structure, **kwargs):
    '''Show pytraj trajectory.

    Examples
    --------
    >>> import nglview as nv
    >>> import parmed as pmd
    >>> t = pmd.load_file(nv.datafiles.PDB)
    >>> w = nv.show_parmed(t)
    >>> w # doctest: +SKIP
    '''
    structure_trajectory = ParmEdTrajectory(parmed_structure)
    return NGLWidget(structure_trajectory, **kwargs)


def show_rdkit(rdkit_mol, **kwargs):
    '''Show rdkit's Mol.

    Parameters
    ----------
    rdkit_mol : rdkit.Chem.rdchem.Mol
    kwargs : additional keyword argument

    Examples
    --------
    >>> import nglview as nv
    >>> from rdkit import Chem # doctest: +SKIP
    ... from rdkit.Chem import AllChem
    ... m = Chem.AddHs(Chem.MolFromSmiles('COc1ccc2[C@H](O)[C@@H](COc2c1)N3CCC(O)(CC3)c4ccc(F)cc4'))
    ... _ = AllChem.EmbedMultipleConfs(m, useExpTorsionAnglePrefs=True, useBasicKnowledge=True)
    ... view = nv.show_rdkit(m)
    ... view # doctest: +SKIP

    >>> # add component m2
    >>> # create file-like object
    >>> from nglview.show import StringIO
    >>> m2 = Chem.AddHs(Chem.MolFromSmiles('N[C@H](C)C(=O)O')) # doctest: +SKIP
    ... fh = StringIO(Chem.MolToPDBBlock(m2))
    ... view.add_component(fh, ext='pdb')

    >>> # load as trajectory, need to have ParmEd
    >>> view = nv.show_rdkit(m, parmed=True) # doctest: +SKIP
    '''
    from rdkit import Chem
    fh = StringIO(Chem.MolToPDBBlock(rdkit_mol))

    try:
        use_parmed = kwargs.pop("parmed")
    except KeyError:
        use_parmed = False

    if not use_parmed:
        view = NGLWidget()
        view.add_component(fh, ext='pdb', **kwargs)
        return view
    else:
        import parmed as pmd
        parm = pmd.load_rdkit(rdkit_mol)
        parm_nv = ParmEdTrajectory(parm)

        # set option for ParmEd
        parm_nv.only_save_1st_model = False

        # set option for NGL
        # wait for: https://github.com/arose/ngl/issues/126
        # to be fixed in NGLView
        # parm_nv.params = dict(firstModelOnly=True)
        return NGLWidget(parm_nv, **kwargs)


def show_mdanalysis(atomgroup, **kwargs):
    '''Show NGL widget with MDAnalysis AtomGroup.

    Can take either a Universe or AtomGroup as its data input.

    Examples
    --------
    >>> import nglview as nv # doctest: +SKIP
    ... import MDAnalysis as mda
    ... u = mda.Universe(nv.datafiles.GRO, nv.datafiles.XTC)
    ... prot = u.select_atoms('protein')
    ... w = nv.show_mdanalysis(prot)
    ... w
    '''
    structure_trajectory = MDAnalysisTrajectory(atomgroup)
    return NGLWidget(structure_trajectory, **kwargs)


def show_htmd(mol, **kwargs):
    '''Show NGL widget with HTMD Molecule.

    Takes a Molecule object as its data input.

    Examples
    --------
    >>> import nglview as nv # doctest: +SKIP
    ... from htmd import Molecule
    ... mol = Molecule(nv.datafiles.PDB)
    ... mol.filter('protein')
    ... w = nv.show_htmd(mol)
    ... w
    '''
    structure_trajectory = HTMDTrajectory(mol)
    return NGLWidget(structure_trajectory, **kwargs)


def show_schrodinger(mol, traj=None, **kwargs):
    '''Show NGL widget with Schrodinger's Structure

    Notes
    -----
    EXPERIMENTAL

    Examples
    --------
    >>> import nglview as nv # doctest: +SKIP
    ... from schrodinger.structure import StructureReader
    ... for s in StructureReader(fn):
    ...    break
    ... w = nv.show_schrodinger(s)
    ... w
    '''
    if traj is None:
        structure_trajectory = SchrodingerStructure(mol)
    else:
        structure_trajectory = SchrodingerTrajectory(mol, traj)
    return NGLWidget(structure_trajectory, **kwargs)


def show_biopython(entity, **kwargs):
    '''Show NGL widget with Biopython structural entity.

    Takes a Structure, Model, Chain, Residue or Atom
    from Bio.PDB as its data input.

    Examples
    --------
    >>> import nglview as nv # doctest: +SKIP
    ... from Bio.PDB import PDBParser
    ... parser = PDBParser()
    ... structure = parser.get_structure("protein", "protein.pdb")
    ... w = nv.show_biopython(structure[0]["A"])
    ... w
    '''
    entity = BiopythonStructure(entity)
    return NGLWidget(entity, **kwargs)


def demo(*args, **kwargs):
    '''

    Examples
    --------
    >>> import nglview as nv
    >>> view = nv.demo()
    >>> view # doctest: +SKIP
    '''
    from nglview import show_structure_file
    return show_structure_file(datafiles.PDB, *args, **kwargs)



import subprocess
import sys

import numpy as np
from mock import MagicMock

import nglview
from make_dummy_comm import *

# TODO : add more show_xxx
# (check test_widgets.py)


def _write(*args, **kargs):
    # fake write method
    subprocess.check_call(['cp', nglview.datafiles.PDB, 'tmp.pdb'])


class MockStructure:
    def as_pdb_string(self):
        with open(nglview.datafiles.PDB) as fh:
            return fh.read()


class MockRosettaPose:
    def dump_pdb(self, _):
        _write()


def test_show_schrodinger():
    # Show a schrodinger.structure.Structure
    s = MagicMock()
    s.write = _write
    v0 = nglview.show_schrodinger(s)

    # Show a trajectory with a Structure as topology
    class MockAtom:
        pdbname = 'C'
        atomic_number = 6
        pdbres = 'GLU'
        resnum = 1
        chain = 'A'

    s = MagicMock()
    s.fsys_ct.getXYZ.return_value = np.zeros((3, 3))
    s.fsys_ct.atom = [MockAtom()] * 3
    traj = MagicMock()
    traj.__len__.return_value = 3
    v1 = nglview.show_schrodinger(s, traj)
    assert v1.max_frame == 2
    v1.frame = 2  # trigger `get_coordinates`


def test_show_htmd():
    mol = MagicMock()
    mol.write = _write
    n_frames = 10
    mol.coordinates = np.zeros((n_frames, 1000, 3))
    mol.numFrames = n_frames
    view = nglview.show_htmd(mol)
    view


def test_show_rosetta():
    pose = MockRosettaPose()
    view = nglview.show_rosetta(pose)


def test_show_iotbx():
    mol = MockStructure()
    nglview.show_iotbx(mol)


def test_show_iodata():
    class MockIO:
        def to_file(self, fname):
            with open(fname, 'w') as fh:
                fh.write("""3
water
H    0.7838370000   -0.4922360000   -0.0000000000
O   -0.0000000000    0.0620200000   -0.0000000000
H   -0.7838370000   -0.4922360000   -0.0000000000""")

    v = nglview.show_iodata(MockIO())
    v


def test_show_qcelemental_show_psi4():
    class MockMol:
        def to_string(self, format):
            return '1\nHe\nHe                    0.000000000000     0.000000000000     0.000000000000\n'

    v = nglview.show_qcelemental(MockMol())
    v
    nglview.show_psi4(MockMol())


def test_show_openbabel():
    import types
    openbabel = types.ModuleType('openbabel')
    sys.modules['openbabel'] = openbabel
    b = MagicMock()
    mol = MagicMock()
    openbabel.OBConversion = MagicMock(return_value=b)
    nglview.show_openbabel(mol)
    b.SetOutFormat.assert_called_with('pdb')
    assert b.WriteFile.called
    assert len(b.WriteFile.call_args_list[0]) == 2


def test_show_prody():
    import types
    prody = types.ModuleType('prody')
    sys.modules['prody'] = prody

    class MockEnsemble:
        def __getitem__(self, index):
            return 0

        def numConfs(self):
            return 1

        def getConformation(self, index):
            class Struct:
                def getCoords(self):
                    return

    prody.Ensemble = MockEnsemble
    prody.writePDB = MagicMock()
    ens = MockEnsemble()
    nglview.show_prody(ens)
    assert prody.writePDB.called
    prody.writePDB.reset_mock()
    assert not prody.writePDB.called

    st = MagicMock()
    nglview.show_prody(st)
    assert prody.writePDB.called



import os

from mock import MagicMock, patch

import nglview
import pytraj
from make_dummy_comm import *

# local


class FakeEvent:
    def is_set(self):
        return self._event_set


@patch('moviepy.editor.ImageSequenceClip')
def test_movie_maker(ImageSequenceClip):
    from nglview.contrib.movie import MovieMaker
    ImageSequenceClip.write_gif = MagicMock()
    ImageSequenceClip.write_videofile = MagicMock()
    traj = pytraj.datafiles.load_tz2()
    view = nglview.show_pytraj(traj)

    movie = MovieMaker(view, in_memory=False)
    movie.download_folder = os.path.join(os.path.dirname(__file__), 'data')

    # fake _event
    movie._event = FakeEvent()
    movie._event._event_set = True
    movie.make()

    movie._event._event_set = False
    movie.make()

    movie = MovieMaker(view, in_memory=False)
    movie.skip_render = True
    movie.make()

    movie.output = 'hello.mp4'
    movie.make()

    movie.interupt()
    movie._event = None
    movie.interupt()

    movie.in_memory = True
    movie.make()

    movie._event._event_set = False
    movie.make()

    movie = MovieMaker(view,
                       download_folder='here',
                       render_params=dict(factor=4),
                       moviepy_params={},
                       stop=2)
    movie.make()


def test_movie_maker_base64_to_ndarray():
    from nglview.contrib.movie import MovieMaker
    s = 'iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO9TXL0Y4OHwAAAABJRU5ErkJggg=='
    MovieMaker._base64_to_ndarray(s)



from .utils import js_utils


class DatasourceRegistry:
    sources = {}

    @classmethod
    def add(cls, name, url):
        """

        Examples
        --------
        >>> import nglview as nv
        >>> from nglview.data_source import DatasourceRegistry # doctest: +SKIP
        ... DatasourceRegistry.add("data", "//cdn.rawgit.com/arose/ngl/v2.0.0-dev.32/data/")
        ... view = nv.NGLWidget()
        ... v.add_component('data://1CRN.cif') # 1CRN.cif is from above url
        ... v
        """
        cls.sources[name] = url

        js_utils.run(f"""
        var NGL = require("nglview-js-widgets").NGL;

        NGL.DatasourceRegistry.add("{name}",
            new NGL.StaticDatasource("{url}"))
        """)



from uuid import uuid4

from ipywidgets import Button, GridBox, HBox, Layout  # ipywidgets >= 7.3

from nglview import NGLWidget
from nglview.utils import js_utils

_code_set_size = """
var ww = window.outerWidth
var wh = window.outerHeight
var vw = ww / %s + 'px'
var vh = wh / %s + 50 + 'px'
var pmodel =  this.model.widget_manager.get_model('%s')

pmodel.then(function(model){
    var key = Object.keys(model.views)[0];
    model.views[key].then(function(v){
        v.children_views.views.forEach(function(pv){
             pv.then(function(v){
                 v.setSize(vw, vh);
             })
        })
    })
})
"""

_code_esc_callback = """
document.onkeydown = function(event){
    if (event.keyCode === 27){
        pmodel.then(function(model){
            var key = Object.keys(model.views)[0];
            model.views[key].then(function(v){
                v.children_views.views.forEach(function(pv){
                    pv.then(function(pvv){
                        pvv.setSize('400px', '300px');
                    })
                })
            })
        })
    }
}
"""


class GridBoxNGL(GridBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._ngl_class_id = None

    def add_class(self, class_id):
        super().add_class(class_id)
        self._ngl_class_id = class_id
        return self

    def fullscreen(self, js_code=None):
        code = (f"""
            var ele = document.getElementsByClassName('{self._ngl_class_id}')[0];
            ele.requestFullscreen();
        """)
        n_rows = len(self.children) // self._n_columns
        n_columns = self._n_columns
        code_fullscreen = _code_set_size % (n_columns, n_rows,
                                            self.model_id) + _code_esc_callback
        self.children[0]._execute_js_code(code_fullscreen)

        if js_code is not None:
            code += js_code
        js_utils.run(code)


class GridBoxViewAndPlayer(GridBoxNGL):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._view = self.children[0]
        self._player = self._view.player
        self._player._create_all_widgets()
        self._add_button_fullscreen()

    def _add_button_fullscreen(self):
        btn = Button(description='Fullscreen', icon='fullscreen')

        def on_click(b):
            self.fullscreen()

        btn.on_click(on_click)

        self._player.widget_general.children = list(
            self._player.widget_general.children) + [btn]


# def _sync_camera_pair(v0, v1):
#     v0._set_sync_camera(v0
#     def on_change_camera(change):
#         new = change['new']
#         owner = change['owner']
#         if owner == v0 and v0._ngl_focused:
#             v1.control.orient(v0._camera_orientation)
#         elif owner == v1 and v1._ngl_focused:
#             v0.control.orient(v1._camera_orientation)
#
#     v0.observe(on_change_camera, '_camera_orientation')
#     v1.observe(on_change_camera, '_camera_orientation')

# def _sync_all(views):
#     from itertools import combinations
#     views = [v for v in views if isinstance(v, NGLWidget)]
#     for v0, v1 in combinations(views, 2):
#         _sync_camera_pair(v0, v1)


def _sync_all(views):
    views = {v for v in views if isinstance(v, NGLWidget)}
    for v in views:
        v._set_sync_camera(views)


def grid_view(views,
              n_columns,
              grid_class=GridBoxViewAndPlayer,
              fullscreen=False,
              sync_camera=False,
              **grid_kwargs):
    if sync_camera:
        _sync_all(views)

    grid_template_columns = f"{str(100 / n_columns)}% " * n_columns
    grid_kwargs = grid_kwargs or {
        'grid_template_columns': grid_template_columns,
        'grid_spacing': '0px 0px'
    }
    box = grid_class(views, layout=Layout(**grid_kwargs))
    class_id = f"nglview-grid-{str(uuid4())}"
    box.add_class(class_id)
    box._n_columns = n_columns

    if fullscreen:

        def on_displayed(box):
            box.fullscreen()

        box.on_displayed(on_displayed)
    return box


def fullscreen_mode(view):
    player = view.player
    player._create_all_widgets()
    b = player._display()
    b.layout.width = '400px'
    b.layout.align_self = 'stretch'
    bb = GridBox([view, b], layout=Layout(grid_template_columns='70% 30%'))
    class_id = f'nglview-{uuid4()}'
    bb.add_class(class_id)

    btn = Button(description='Fullscreen', icon='fullscreen')

    def on_click(b):
        view._execute_js_code("""
        var ele = document.getElementsByClassName("%s")[0];

        this.stage.viewer.container.style.height = window.screen.height + 'px'
        console.log(window.innerHeight)
        this.stage.toggleFullscreen(ele);
        """ % class_id)

    btn.on_click(on_click)
    player.widget_general.children = list(
        player.widget_general.children) + [btn]
    return bb



import socket
from itertools import chain

from notebook.notebookapp import NotebookApp, random_ports
from tornado import httpserver


class NGLViewApp(NotebookApp):
    def get_port(self, port=None):

        # user-given port
        uport = [
            port,
        ] if port is not None else [
            self.port,
        ]

        def func():
            return

        self.http_server = httpserver.HTTPServer(func)
        success = None

        for port in chain(uport, random_ports(self.port,
                                              self.port_retries + 1)):
            try:
                self.http_server.listen(port, self.ip)
            except (OSError, socket.error):
                pass
            else:
                self.port = port
                success = True
                break
        if not success:
            self.log.critical(
                'ERROR: the notebook server could not be started because '
                'no available port could be found.')
            self.exit(1)
        self.http_server.stop()
        return self.port



from ipywidgets import Box, interactive
from traitlets import Any, Bool, Dict, Int, observe

# local
from .color import COLOR_SCHEMES
from .utils import py_utils


class RepresentationControl(Box):
    parameters = Dict().tag(sync=False)
    name = Any().tag(sync=False)
    repr_index = Int().tag(sync=False)
    component_index = Int().tag(sync=False)
    _disable_update_parameters = Bool(False).tag(sync=False)

    def __init__(self,
                 view,
                 component_index,
                 repr_index,
                 name=None,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.component_index = component_index
        self.repr_index = repr_index
        self._view = view
        self.children = self._make_widget().children
        # trigger
        self.name = name

    def _on_change_widget_child_value(self, change):
        owner = change['owner']
        new = change['new']
        self.parameters = {py_utils._camelize(owner._ngl_description): new}

    @observe('parameters')
    def _on_parameters_changed(self, change):
        if not self._disable_update_parameters:
            parameters = change['new']

            self._view.update_representation(component=self.component_index,
                                             repr_index=self.repr_index,
                                             **parameters)

    @observe('name')
    def _on_name_changed(self, change):
        new_name = change['new']
        if new_name == 'surface':
            for kid in self.children:
                if kid._ngl_type == 'surface':
                    kid.layout.display = 'flex'
        else:
            for kid in self.children:
                if kid._ngl_type == 'surface':
                    kid.layout.display = 'none'

    def _get_name_and_repr_dict(self, c_string, r_string):
        try:
            _repr_dict = self._view._ngl_repr_dict[c_string][r_string][
                'parameters']
            name = self._view._ngl_repr_dict[c_string][r_string]['name']
        except KeyError:
            _repr_dict = dict()
            name = ''

        return name, _repr_dict

    @observe('repr_index')
    def _on_repr_index_changed(self, change):
        c_string = 'c' + str(self.component_index)
        r_string = str(change['new'])
        self._update(c_string, r_string)

    @observe('component_index')
    def _on_component_index_changed(self, change):
        c_string = 'c' + str(change['new'])
        r_string = str(self.repr_index)
        self._update(c_string, r_string)

    def _update(self, c_string, r_string):
        name, _repr_dict = self._get_name_and_repr_dict(c_string, r_string)
        self.name = name
        self._disable_update_parameters = True
        for kid in self.children:
            desc = py_utils._camelize(kid._ngl_description)
            if desc in _repr_dict:
                kid.value = _repr_dict.get(desc)
        self._disable_update_parameters = False

    def _make_widget(self):
        c_string = 'c' + str(self.component_index)
        r_string = str(self.repr_index)
        name, _repr_dict = self._get_name_and_repr_dict(c_string, r_string)

        assembly_list = ['default', 'AU', 'BU1', 'UNITCELL', 'SUPERCELL']
        surface_types = ['vws', 'sas', 'ms', 'ses']

        def func(opacity=_repr_dict.get('opacity', 1.),
                 assembly=_repr_dict.get('assembly', 'default'),
                 color_scheme=_repr_dict.get('colorScheme', " "),
                 wireframe=_repr_dict.get('wireframe', False),
                 probe_radius=_repr_dict.get('probeRadius', 1.4),
                 isolevel=_repr_dict.get('isolevel', 2.),
                 smooth=_repr_dict.get('smooth', 2.),
                 surface_type=_repr_dict.get('surfaceType', 'ms'),
                 box_size=_repr_dict.get('boxSize', 10),
                 cutoff=_repr_dict.get('cutoff', 0)):
            pass

        widget = interactive(func,
                             opacity=(0., 1., 0.1),
                             color_scheme=COLOR_SCHEMES,
                             assembly=assembly_list,
                             probe_radius=(0., 5., 0.1),
                             isolevel=(0., 10., 0.1),
                             smooth=(0, 10, 1),
                             surface_type=surface_types,
                             box_size=(0, 100, 2),
                             cutoff=(0., 100, 0.1),
                             continuous_update=False)
        for kid in widget.children:
            try:
                setattr(kid, '_ngl_description', kid.description)
            except AttributeError:
                # ipywidgets.Output does not have `description` attribute
                setattr(kid, '_ngl_description', '')
            if kid._ngl_description in [
                    'probe_radius', 'smooth', 'surface_type', 'box_size',
                    'cutoff'
            ]:
                setattr(kid, '_ngl_type', 'surface')
            else:
                setattr(kid, '_ngl_type', 'basic')
            kid.observe(self._on_change_widget_child_value, 'value')
        return widget



from functools import partial

from IPython.display import HTML, Javascript, display

__alll__ = [
    'clean_error_output',
    'launch_qtconsole',
    'clean_empty_output_area',
    'open_url_template',
    '_set_ipython_cell',
    'ngl_demo',
    'init_funcs',
]


def run(command):
    display(Javascript(command))


# cell.clear_output()
def hide_toolbar():
    run("$('#maintoolbar').hide()")
    run("$('#header-container').hide()")


def show_toolbar():
    run("$('#maintoolbar').show()")
    run("$('#header-container').show()")


def execute(command):
    run(f'Jupyter.notebook.kernel.execute("{command}")')


def _set_notebook_width(width='20%', left_padding=0):
    script_template = """
    var cb = Jupyter.notebook.container;

    cb.width('{width}');

    """.format(width=width)

    if left_padding is not None:
        offset_str = f"cb.offset({{'left': {left_padding}}})"
    else:
        offset_str = ''
    command = script_template + offset_str
    run(command)


clean_empty_output_area = partial(run,
                                  command="""
var output_area = $(".output_area");

for (var i=0; i < output_area.length; i++){
    if (!output_area[i].innerText){
        output_area[i].remove();
    }
}
""")

launch_qtconsole = partial(run,
                           command="""
Jupyter.notebook.kernel.execute('%qtconsole')
""")

open_url_template = """
window.open({url});
"""

clean_error_output = partial(run,
                             command="""
var cells = Jupyter.notebook.get_cells();

for (var i = 0; i < cells.length; i++){
    var cell = cells[i];
    if (cell.output_area.outputs.length > 0) {
        var out = cell.output_area.outputs[0];
        if (out.output_type == 'error') {
            cell.clear_output();
        }
    }
}
""")


def _set_ipython_cell(background='#67a9cf'):
    cm = """
    var setIPythonLikeCell = function(){
        //var cell = Jupyter.notebook.insert_cell_at_bottom();
        var cell = Jupyter.notebook.get_selected_cell();
        var $el = cell.element;
        cell.set_text('');
        $el.css({'background': '%s'});

        var handler = function(event) {
            var selected_cell = Jupyter.notebook.get_selected_cell();
            if (selected_cell.cell_id === cell.cell_id){
                selected_cell.execute();
                selected_cell.set_text('');
            }
            return false;
        };

        var action = {
            help: 'run cell',
            help_index: 'zz',
            handler: handler
        };

        Jupyter.keyboard_manager.edit_shortcuts.add_shortcut('enter', action); 
    };

    setIPythonLikeCell()
    """ % background

    cm2 = """
    var cell = Jupyter.notebook.get_selected_cell();
    cell.element.draggable().resizable();
    """
    run(cm)
    run(cm2)


def ngl_demo(width=400, height=400):
    """make a viewport, create stage object and populate NGL namespace
    """
    command = """
    <div id='viewport'></div>
    <script>
        $('#viewport').width(%s).height(%s);
    </script>
    """ % (width, height)

    command2 = """
    <script>
        var NGL = require('nbextensions/nglview-js-widgets/index').NGL;
        var stage = new NGL.Stage('viewport')
    </script>
    """

    display(HTML(command))
    display(HTML(command2))


def init_funcs():
    """print
    """
    command = """
    <script>
        var print = function(x){
            for (var i in x){
                console.log(i)
            }
        }
    </script>
    """
    display(HTML(command))



# TODO: reorg
# simplify code
import json
import time
import uuid
from collections import defaultdict

from IPython.display import Javascript, display
from ipywidgets import (Box, Button, Checkbox, ColorPicker, Dropdown,
                        FloatSlider, FloatText, HBox, IntSlider, IntText,
                        Label, Layout, Play, Tab, Text, Textarea, ToggleButton,
                        VBox, interactive, jslink)
from traitlets import (Any, Bool, CaselessStrEnum, Dict, Float, HasTraits, Int,
                       link, observe)
import traitlets

from . import default
from .layout import (_make_autofit, _make_box_layout, _make_delay_tab,
                     _relayout, _relayout_master, make_form_item_layout)
from .parameters import REPRESENTATION_NAMES
from .utils import js_utils


def _dry_run(v, func, *args, **kwargs):
    # Get remote call message.
    msg = []

    def _another_call(*args, **kwargs):
        msg.append(v._get_remote_call_msg(*args, **kwargs))

    old = v._remote_call
    setattr(v, '_remote_call', _another_call)
    func(*args, **kwargs)
    setattr(v, '_remote_call', old)

    return msg and msg[-1] or None  # FIXME: all messages?


def _cature_msg(f):
    def wrap(*args, **kwargs):
        self = args[0]
        v = self._view
        msg_dict = v._player_dict
        w = f(*args, **kwargs)
        # FIXME: add code here
        # if isinstance(w, Button):
        #     msg_dict[w.model_id] = _dry_run(v, w.click)
        return w
    return wrap


class TrajectoryPlayer(HasTraits):
    # should set default values here different from desired defaults
    # so `observe` can be triggered
    step = Int(0)
    interpolate = Bool(False)
    delay = Float(0.0)
    parameters = Dict()
    iparams = Dict()
    _interpolation_t = Float()
    _iterpolation_type = CaselessStrEnum(['linear', 'spline'])
    camera = CaselessStrEnum(['perspective', 'orthographic'],
                             default_value='perspective')
    _render_params = Dict()
    _real_time_update = Bool(False)

    widget_tab = Any(None)
    widget_repr = Any(None)
    widget_repr_parameters = Any(None)
    widget_quick_repr = Any(None)
    widget_general = Any(None)
    widget_picked = Any(None)
    widget_preference = Any(None)
    widget_extra = Any(None)
    widget_help = Any(None)
    widget_export_image = Any(None)
    widget_component_slider = Any(None)
    widget_repr_slider = Any(None)
    widget_repr_choices = Any(None)
    widget_repr_control_buttons = Any(None)
    widget_repr_add = Any(None)
    widget_accordion_repr_parameters = Any(None)
    widget_repr_parameters_dialog = Any(None)
    widget_repr_name = Any(None)
    widget_component_dropdown = Any(None)
    widget_player = Any(None)
    widget_background = Any(None)
    widget_camera = Any(None)
    btn_center = Any(None)

    def __init__(self, view, step=1, delay=100, min_delay=40):
        self._view = view
        self.step = step
        self.delay = delay
        self.min_delay = min_delay
        self._iplayer = None
        self._interpolation_t = 0.5
        self._iterpolation_type = 'linear'
        self.iparams = dict(t=self._interpolation_t,
                            step=1,
                            type=self._iterpolation_type)
        self._render_params = dict(factor=4,
                                   antialias=True,
                                   trim=False,
                                   transparent=False)

        self._widget_names = [w for w in dir(self) if w.startswith('wiget_')]

    def _update_padding(self, padding=default.DEFAULT_PADDING):
        widget_collection = [
            self.widget_general, self.widget_repr, self.widget_preference,
            self.widget_repr_parameters, self.widget_help, self.widget_extra,
            self.widget_picked
        ]
        for widget in widget_collection:
            if widget is not None:
                widget.layout.padding = padding

    def _create_all_widgets(self, thread=False):

        def make():
            self.widget_tab = self._display()

            old_index = self.widget_tab.selected_index
            for index, _ in enumerate(self.widget_tab.children):
                self.widget_tab.selected_index = index
                if thread:
                    time.sleep(0.1)

            self.widget_tab.selected_index = old_index
        if thread:
            self._view._run_on_another_thread(make)
        else:
            make()

    def smooth(self):
        self.interpolate = True

    @observe('delay')
    def _on_delay(self, change):
        if self.widget_player:
            self.widget_player.interval = change['new']

    @observe('camera')
    def on_camera_changed(self, change):
        camera_type = change['new']
        self._view._remote_call("setParameters",
                                target='Stage',
                                kwargs=dict(cameraType=camera_type))

    @property
    def frame(self):
        return self._view.frame

    @frame.setter
    def frame(self, value):
        self._view.frame = value

    @property
    def max_frame(self):
        return self._view.max_frame

    @observe('parameters')
    def update_parameters(self, change):
        params = change['new']
        self.delay = params.get("delay", self.delay)
        self.step = params.get("step", self.step)

    @observe('_interpolation_t')
    def _interpolation_t_changed(self, change):
        self.iparams['t'] = change['new']

    def _display(self):
        if self.widget_tab is None:
            box_factory = [(self._make_general_box, 'General'),
                           (self._make_widget_repr, 'Representation'),
                           (self._make_widget_preference, 'Preference'),
                           (self._make_extra_box, 'Extra')]

            tab = _make_delay_tab(box_factory, selected_index=0)
            # tab = _make_autofit(tab)
            tab.layout.align_self = 'center'
            tab.layout.align_items = 'stretch'

            self.widget_tab = self._view._igui = tab

        return self.widget_tab

    def _make_widget_tab(self):
        return self._display()

    @_cature_msg
    def _make_button_center(self):
        self.btn_center = Button(description=' Center', icon='fa-bullseye')

        @self.btn_center.on_click
        def on_click(_):
            self._view.center()

        return self.btn_center

    def _make_widget_preference(self, width='100%'):
        def make_func():
            parameters = self._view._ngl_full_stage_parameters

            def func(pan_speed=parameters.get('panSpeed', 0.8),
                     rotate_speed=parameters.get('rotateSpeed', 2),
                     zoom_speed=parameters.get('zoomSpeed', 1.2),
                     clip_dist=parameters.get('clipDist', 10),
                     camera_fov=parameters.get('cameraFov', 40),
                     clip_far=parameters.get('clipFar', 100),
                     clip_near=parameters.get('clipNear', 0),
                     fog_far=parameters.get('fogFar', 100),
                     fog_near=parameters.get('fogNear', 50),
                     impostor=parameters.get('impostor', True),
                     light_intensity=parameters.get('lightIntensity', 1),
                     quality=parameters.get('quality', 'medium'),
                     sample_level=parameters.get('sampleLevel', 1)):

                self._view.parameters = dict(panSpeed=pan_speed,
                                             rotateSpeed=rotate_speed,
                                             zoomSpeed=zoom_speed,
                                             clipDist=clip_dist,
                                             clipFar=clip_far,
                                             clipNear=clip_near,
                                             cameraFov=camera_fov,
                                             fogFar=fog_far,
                                             fogNear=fog_near,
                                             impostor=impostor,
                                             lightIntensity=light_intensity,
                                             quality=quality,
                                             sampleLevel=sample_level)

            return func

        def make_widget_box():
            widget_sliders = interactive(make_func(),
                                         pan_speed=(0, 10, 0.1),
                                         rotate_speed=(0, 10, 1),
                                         zoom_speed=(0, 10, 1),
                                         clip_dist=(0, 200, 5),
                                         clip_far=(0, 100, 1),
                                         clip_near=(0, 100, 1),
                                         camera_fov=(15, 120, 1),
                                         fog_far=(0, 100, 1),
                                         fog_near=(0, 100, 1),
                                         light_intensity=(0, 10, 0.02),
                                         quality=['low', 'medium', 'high'],
                                         sample_level=(-1, 5, 1))

            for child in widget_sliders.children:
                if isinstance(child, (IntSlider, FloatSlider)):
                    child.layout.width = default.DEFAULT_SLIDER_WIDTH
            return widget_sliders

        if self.widget_preference is None:
            widget_sliders = make_widget_box()
            reset_button = Button(description='Reset')
            widget_sliders.children = [
                reset_button,
            ] + list(widget_sliders.children)

            @reset_button.on_click
            def on_click(reset_button):
                self._view.parameters = self._view._original_stage_parameters
                self._view._ngl_full_stage_parameters = self._view._original_stage_parameters
                widget_sliders.children = [
                    reset_button,
                ] + list(make_widget_box().children)

            self.widget_preference = _relayout_master(widget_sliders,
                                                      width=width)
        return self.widget_preference

    def _show_download_image(self):
        # "interactive" does not work for True/False in ipywidgets 4 yet.
        button = Button(description=' Screenshot', icon='fa-camera')

        @button.on_click
        def on_click(button):
            self._view.download_image()

        return button

    def _make_text_picked(self):
        ta = Textarea(value=json.dumps(self._view.picked),
                      description='Picked atom')
        ta.layout.width = '300px'
        _dry_run(self._view,
                 self._view._on_picked,
                 change={
                     'old': '',
                     'new': ''
                 })
        return ta

    def _refresh(self, component_slider, repr_slider):
        """update representation and component information
        """
        self._view._request_repr_parameters(component=component_slider.value,
                                            repr_index=repr_slider.value)
        self._view._update_repr_dict()
        self._view._handle_repr_dict_changed(change=dict(
            new=self._view._ngl_repr_dict))

    @_cature_msg
    def _make_button_repr_control(self, component_slider, repr_slider,
                                  repr_selection):
        button_refresh = Button(description=' Refresh',
                                tooltip='Get representation info',
                                icon='fa-refresh')
        button_center_selection = Button(description=' Center',
                                         tooltip='center selected atoms',
                                         icon='fa-bullseye')
        button_center_selection._ngl_name = 'button_center_selection'
        button_hide = Button(description=' Hide',
                             icon='fa-eye-slash',
                             tooltip='Hide/Show current representation')
        button_remove = Button(description=' Remove',
                               icon='fa-trash',
                               tooltip='Remove current representation')

        @button_refresh.on_click
        def on_click_refresh(button):
            self._refresh(component_slider, repr_slider)

        @button_center_selection.on_click
        def on_click_center(center_selection):
            self._view.center(selection=repr_selection.value,
                              component=component_slider.value)

        @button_hide.on_click
        def on_click_hide(button_hide):
            component = component_slider.value
            repr_index = repr_slider.value

            if button_hide.description == 'Hide':
                hide = True
                button_hide.description = 'Show'
            else:
                hide = False
                button_hide.description = 'Hide'

            self._view._remote_call('setVisibilityForRepr',
                                    target='Widget',
                                    args=[component, repr_index, not hide])

        @button_remove.on_click
        def on_click_remove(button_remove):
            self._view._remove_representation(component=component_slider.value,
                                              repr_index=repr_slider.value)
            self._view._request_repr_parameters(
                component=component_slider.value, repr_index=repr_slider.value)

        bbox = _make_autofit(
            HBox([
                button_refresh, button_center_selection, button_hide,
                button_remove
            ]))
        for b in bbox.children:
            if hasattr(b, 'click'):
                _dry_run(self._view, b.click)
        return bbox

    def _make_widget_repr(self):
        self.widget_repr_name = Text(value='', description='representation')
        self.widget_repr_name._ngl_name = 'repr_name_text'
        repr_selection = Text(value=' ', description='selection')
        repr_selection._ngl_name = 'repr_selection'
        repr_selection.width = self.widget_repr_name.width = default.DEFAULT_TEXT_WIDTH

        max_n_components = max(self._view.n_components - 1, 0)
        self.widget_component_slider = IntSlider(value=0,
                                                 max=max_n_components,
                                                 min=0,
                                                 description='component')
        self.widget_component_slider._ngl_name = 'component_slider'

        cvalue = ' '
        self.widget_component_dropdown = Dropdown(value=cvalue,
                                                  options=[
                                                      cvalue,
                                                  ],
                                                  description='component')
        self.widget_component_dropdown._ngl_name = 'component_dropdown'

        self.widget_repr_slider = IntSlider(value=0,
                                            description='representation',
                                            width=default.DEFAULT_SLIDER_WIDTH)
        self.widget_repr_slider._ngl_name = 'repr_slider'
        self.widget_repr_slider.visible = True

        self.widget_component_slider.layout.width = default.DEFAULT_SLIDER_WIDTH
        self.widget_repr_slider.layout.width = default.DEFAULT_SLIDER_WIDTH
        self.widget_component_dropdown.layout.width = self.widget_component_dropdown.max_width = default.DEFAULT_TEXT_WIDTH

        # turn off for now
        self.widget_component_dropdown.layout.display = 'none'
        self.widget_component_dropdown.description = ''

        # self.widget_accordion_repr_parameters = Accordion()
        self.widget_accordion_repr_parameters = Tab()
        self.widget_repr_parameters = self._make_widget_repr_parameters(
            self.widget_component_slider, self.widget_repr_slider,
            self.widget_repr_name)
        self.widget_accordion_repr_parameters.children = [
            self.widget_repr_parameters, Box()
        ]
        self.widget_accordion_repr_parameters.set_title(0, 'Parameters')
        self.widget_accordion_repr_parameters.set_title(1, 'Hide')
        self.widget_accordion_repr_parameters.selected_index = 1

        checkbox_reprlist = Checkbox(value=False, description='reprlist')
        checkbox_reprlist._ngl_name = 'checkbox_reprlist'
        self.widget_repr_choices = self._make_repr_name_choices(
            self.widget_component_slider, self.widget_repr_slider)
        self.widget_repr_choices._ngl_name = 'reprlist_choices'

        self.widget_repr_add = self._make_add_widget_repr(
            self.widget_component_slider)

        def on_update_checkbox_reprlist(change):
            self.widget_repr_choices.visible = change['new']

        checkbox_reprlist.observe(on_update_checkbox_reprlist, names='value')

        def on_repr_name_text_value_changed(change):
            name = change['new'].strip()
            old = change['old'].strip()

            should_update = (self._real_time_update and old and name
                             and name in REPRESENTATION_NAMES
                             and name != change['old'].strip())

            if should_update:
                component = self.widget_component_slider.value
                repr_index = self.widget_repr_slider.value
                self._view._remote_call(
                    'setRepresentation',
                    target='Widget',
                    args=[change['new'], {}, component, repr_index])
                self._view._request_repr_parameters(component, repr_index)

        def on_component_or_repr_slider_value_changed(change):
            self._view._request_repr_parameters(
                component=self.widget_component_slider.value,
                repr_index=self.widget_repr_slider.value)
            self.widget_component_dropdown.options = tuple(
                self._view._ngl_component_names)

            if self.widget_accordion_repr_parameters.selected_index >= 0:
                self.widget_repr_parameters.name = self.widget_repr_name.value
                self.widget_repr_parameters.repr_index = self.widget_repr_slider.value
                self.widget_repr_parameters.component_index = self.widget_component_slider.value

        def on_repr_selection_value_changed(change):
            if self._real_time_update:
                component = self.widget_component_slider.value
                repr_index = self.widget_repr_slider.value
                self._view._set_selection(change['new'],
                                          component=component,
                                          repr_index=repr_index)

        def on_change_component_dropdown(change):
            choice = change['new']
            if choice and choice.strip():  # bool(" ") is True
                self.widget_component_slider.value = self._view._ngl_component_names.index(
                    choice)

        self.widget_component_dropdown.observe(on_change_component_dropdown,
                                               names='value')

        self.widget_repr_slider.observe(
            on_component_or_repr_slider_value_changed, names='value')
        self.widget_component_slider.observe(
            on_component_or_repr_slider_value_changed, names='value')
        self.widget_repr_name.observe(on_repr_name_text_value_changed,
                                      names='value')
        repr_selection.observe(on_repr_selection_value_changed, names='value')

        self.widget_repr_control_buttons = self._make_button_repr_control(
            self.widget_component_slider, self.widget_repr_slider,
            repr_selection)

        blank_box = Box([Label("")])

        all_kids = [
            self.widget_repr_control_buttons, blank_box, self.widget_repr_add,
            self.widget_component_dropdown, self.widget_repr_name,
            repr_selection, self.widget_component_slider,
            self.widget_repr_slider, self.widget_repr_choices,
            self.widget_accordion_repr_parameters
        ]

        vbox = VBox(all_kids)

        self._view._request_repr_parameters(
            component=self.widget_component_slider.value,
            repr_index=self.widget_repr_slider.value)

        self.widget_repr = _relayout_master(vbox, width='100%')

        self._refresh(self.widget_component_slider, self.widget_repr_slider)

        setattr(self.widget_repr, "_saved_widgets", [])
        for _box in self.widget_repr.children:
            if hasattr(_box, 'children'):
                for kid in _box.children:
                    self.widget_repr._saved_widgets.append(kid)

        return self.widget_repr

    def _make_widget_repr_parameters(self,
                                     component_slider,
                                     repr_slider,
                                     repr_name_text=None):
        name = repr_name_text.value if repr_name_text is not None else ' '
        widget = self._view._display_repr(component=component_slider.value,
                                          repr_index=repr_slider.value,
                                          name=name)
        widget._ngl_name = 'repr_parameters_box'
        return widget

    def _make_button_export_image(self):
        slider_factor = IntSlider(value=4, min=1, max=10, description='scale')
        checkbox_antialias = Checkbox(value=True, description='antialias')
        checkbox_trim = Checkbox(value=False, description='trim')
        checkbox_transparent = Checkbox(value=False, description='transparent')
        filename_text = Text(value='Screenshot', description='Filename')
        delay_text = FloatText(value=1,
                               description='delay (s)',
                               tooltip='hello')

        start_text, stop_text, step_text = (IntText(value=0,
                                                    description='start'),
                                            IntText(value=self._view.max_frame+1,
                                                    description='stop'),
                                            IntText(value=1,
                                                    description='step'))

        start_text.layout.max_width = stop_text.layout.max_width = step_text.layout.max_width \
                = filename_text.layout.max_width = delay_text.layout.max_width = default.DEFAULT_TEXT_WIDTH

        button_movie_images = Button(description='Export Images')

        def download_image(filename):
            self._view.download_image(factor=slider_factor.value,
                                      antialias=checkbox_antialias.value,
                                      trim=checkbox_trim.value,
                                      transparent=checkbox_transparent.value,
                                      filename=filename)

        @button_movie_images.on_click
        def on_click_images(button_movie_images):
            for i in range(start_text.value, stop_text.value, step_text.value):
                self._view.frame = i
                time.sleep(delay_text.value)
                download_image(filename=filename_text.value + str(i))
                time.sleep(delay_text.value)

        vbox = VBox([
            button_movie_images,
            start_text,
            stop_text,
            step_text,
            delay_text,
            filename_text,
            slider_factor,
            checkbox_antialias,
            checkbox_trim,
            checkbox_transparent,
        ])

        form_items = _relayout(vbox, make_form_item_layout())
        form = Box(form_items, layout=_make_box_layout())
        # form = _relayout_master(vbox)
        return form

    def _make_resize_notebook_slider(self):
        resize_notebook_slider = IntSlider(min=300,
                                           max=2000,
                                           description='resize notebook')

        def on_resize_notebook(change):
            width = change['new']
            self._view._remote_call('resizeNotebook',
                                    target='Widget',
                                    args=[
                                        width,
                                    ])

        resize_notebook_slider.observe(on_resize_notebook, names='value')
        return resize_notebook_slider

    def _make_add_widget_repr(self, component_slider):
        dropdown_repr_name = Dropdown(options=REPRESENTATION_NAMES,
                                      value='cartoon')
        repr_selection = Text(value='*', description='')
        repr_button = Button(description='Add',
                             tooltip="""Add representation.
        You can also hit Enter in selection box""")
        repr_button.layout = Layout(width='auto', flex='1 1 auto')

        dropdown_repr_name.layout.width = repr_selection.layout.width = default.DEFAULT_TEXT_WIDTH

        def on_click_or_submit(button_or_text_area):
            self._view.add_representation(
                selection=repr_selection.value.strip(),
                repr_type=dropdown_repr_name.value,
                component=component_slider.value)

        repr_button.on_click(on_click_or_submit)
        repr_selection.on_submit(on_click_or_submit)
        add_repr_box = HBox([repr_button, dropdown_repr_name, repr_selection])
        add_repr_box._ngl_name = 'add_repr_box'

        return add_repr_box

    def _make_repr_playground(self):
        vbox = VBox()
        children = []

        rep_names = REPRESENTATION_NAMES[:]
        excluded_names = ['ball+stick', 'distance']
        for name in excluded_names:
            rep_names.remove(name)

        repr_selection = Text(value='*')
        repr_selection.layout.width = default.DEFAULT_TEXT_WIDTH
        repr_selection_box = HBox([Label('selection'), repr_selection])
        setattr(repr_selection_box, 'value', repr_selection.value)

        for index, name in enumerate(rep_names):
            button = ToggleButton(description=name)

            def make_func():
                def on_toggle_button_value_change(change, button=button):
                    selection = repr_selection.value
                    new = change['new']  # True/False
                    if new:
                        self._view.add_representation(button.description,
                                                      selection=selection)
                    else:
                        self._view._remove_representations_by_name(
                            button.description)

                return on_toggle_button_value_change

            button.observe(make_func(), names='value')
            children.append(button)

        pd = self._view._player_dict
        pd['widget_quick_repr'] = defaultdict(dict)
        for k in children:
            pd['widget_quick_repr'][k.model_id] = {}

            def click_true(k):
                k.value = True

            pd['widget_quick_repr'][k.model_id][True] = _dry_run(self._view,
                                                                 click_true,
                                                                 k=k)

            def click_false(k):
                k.value = False

            pd['widget_quick_repr'][k.model_id][False] = _dry_run(self._view,
                                                                  click_false,
                                                                  k=k)

        button_clear = Button(description='clear',
                              button_style='info',
                              icon='fa-eraser')

        @button_clear.on_click
        def on_clear(button_clear):
            self._view.clear()
            for kid in children:
                # unselect
                kid.value = False

        if hasattr(button_clear, 'click'):
            _dry_run(self._view, button_clear.click)

        vbox.children = children + [repr_selection, button_clear]
        _make_autofit(vbox)
        self.widget_quick_repr = vbox
        return self.widget_quick_repr

    def _make_repr_name_choices(self, component_slider, repr_slider):
        repr_choices = Dropdown(options=[
            " ",
        ])

        def on_chosen(change):
            repr_name = change.get('new', " ")
            try:
                repr_index = repr_choices.options.index(repr_name)
                repr_slider.value = repr_index
            except ValueError:
                pass

        repr_choices.observe(on_chosen, names='value')
        repr_choices.layout.width = default.DEFAULT_TEXT_WIDTH

        self.widget_repr_choices = repr_choices
        return self.widget_repr_choices

    def _make_widget_picked(self):
        self.widget_picked = self._make_text_picked()
        picked_box = HBox([
            self.widget_picked,
        ])
        return _relayout_master(picked_box, width='75%')

    def _make_export_image_widget(self):
        if self.widget_export_image is None:
            self.widget_export_image = HBox([self._make_button_export_image()])
        return self.widget_export_image

    def _make_extra_box(self):
        if self.widget_extra is None:
            extra_list = [(self._make_widget_picked, 'Picked'),
                          (self._make_repr_playground, 'Quick'),
                          (self._make_export_image_widget, 'Image'),
                          (self._make_command_box, 'Command')]

            self.widget_extra = _make_delay_tab(extra_list, selected_index=0)
        return self.widget_extra

    def _make_general_box(self):
        if self.widget_general is None:
            step_slide = IntSlider(value=self.step,
                                   min=-100,
                                   max=100,
                                   description='step')
            delay_text = IntSlider(value=self.delay,
                                   min=10,
                                   max=1000,
                                   description='delay')
            toggle_button_interpolate = ToggleButton(
                self.interpolate,
                description='Smoothing',
                tooltip='smoothing trajectory')
            link((toggle_button_interpolate, 'value'), (self, 'interpolate'))

            self.widget_background = background_color_picker = ColorPicker(
                value=self._view._ngl_full_stage_parameters.get(
                    'backgroundColor', 'white'),
                description='background')
            self.widget_camera = camera_type = Dropdown(
                value=self.camera,
                options=['perspective', 'orthographic'],
                description='camera')

            link((step_slide, 'value'), (self, 'step'))
            link((delay_text, 'value'), (self, 'delay'))
            link((toggle_button_interpolate, 'value'), (self, 'interpolate'))
            link((camera_type, 'value'), (self, 'camera'))
            link((background_color_picker, 'value'),
                 (self._view, 'background'))

            center_button = self._make_button_center()
            render_button = self._show_download_image()
            center_render_hbox = _make_autofit(
                HBox([
                    toggle_button_interpolate,
                    center_button,
                    render_button,
                ]))

            v0_left = VBox([
                step_slide,
                delay_text,
                background_color_picker,
                camera_type,
                center_render_hbox,
            ])

            v0_left = _relayout_master(v0_left, width='100%')
            self.widget_general = v0_left
        return self.widget_general

    def _ipython_display_(self, **kwargs):
        display(self._display())

    def _make_command_box(self):
        widget_text_command = Text()

        @widget_text_command.on_submit
        def _on_submit_command(_):
            command = widget_text_command.value
            js_utils.execute(command)
            widget_text_command.value = ''

        return widget_text_command

    def _create_all_tabs(self):
        tab = self._display()
        for index, _ in enumerate(tab.children):
            # trigger ceating widgets
            tab.selected_index = index

        self.widget_extra = self._make_extra_box()
        for index, _ in enumerate(self.widget_extra.children):
            self.widget_extra.selected_index = index

    def _simplify_repr_control(self):
        for widget in self.widget_repr._saved_widgets:
            if not isinstance(widget, Tab):
                widget.layout.display = 'none'
        self.widget_repr_choices.layout.display = 'flex'
        self.widget_accordion_repr_parameters.selected_index = 0



class ViewerControl:
    """EXPERIMENTAL. Create viewer controls (rotate, zoom, translation, ...)

    Parameters
    ----------
    view : nglview.NGLWidget

    Notes
    -----
    Unstable feature

    Examples
    --------
    >>> import nglview as nv
    >>> from nglview.viewer_control import ViewerControl
    >>> view = nv.NGLWidget()
    >>> view # doctest: +SKIP
    >>> control = ViewerControl(view=view)
    >>> control.zoom(0.1)
    """

    def __init__(self, view):
        self.view = view

    def _call(self, funcname, *args):
        self.view._remote_call(funcname, target='viewerControls', args=args)

    def _view_xz_plane(self):
        self.view._remote_call('viewXZPlane', target='Widget')

    def align(self, basis):
        '''
        
        Parameters
        ----------
        basis : List[float], len=16
        '''
        self._call('align', basis)

    def apply_matrix(self, basis):
        '''
        
        Parameters
        ----------
        basis : List[float], len=16
        '''
        self._call('applyMatrix', basis)

    def center(self, vector):
        '''
        
        Parameters
        ----------
        vector: List[float], len=3
        '''
        self._call('center', vector)

    def orient(self, basis):
        '''
        
        Parameters
        ----------
        basis : List[float], len=16
        '''
        self._call('orient', basis)

    def rotate(self, basis):
        '''
        
        Parameters
        ----------
        basis : List[float], len=4
            quaternion
        '''
        self._call('rotate', basis)

    def translate(self, vector):
        '''
        
        Parameters
        ----------
        basis : List[float], len=3
        '''
        self._call('translate', vector)

    def spin(self, axis, angle):
        '''
        
        Parameters
        ----------
        axis: List[float], len=3
        angle : float
        '''
        self._call('spin', axis, angle)

    def zoom(self, delta):
        '''
        
        Parameters
        ----------
        delta : float
        '''
        self._call('zoom', delta)



"""highly experimental. Use with your own risk

    >>> from nglview import theme
    >>> theme.oceans16() # oceans16 https://github.com/dunovank/jupyter-themes

Retart your notebook to reset to default Jupyter theme

If you want to set global theme for your notebook, it's better to install jupyter-themes
https://github.com/dunovank/jupyter-themes

"""
import os

style = """
<style id='nglview_style'>
{}
</style>

<script>

var nc = document.getElementById('nglview_style')
document.head.appendChild(nc)
</script>
"""


def _get_theme(css_file):
    from IPython.display import HTML
    return HTML(_get_css_content(css_file))


def _get_css_content(css_file, include_style_tag=True):
    dirname = os.path.dirname(os.path.abspath(__file__))
    css_file = os.path.join(dirname, css_file)
    css = open(css_file).read()
    if include_style_tag:
        return style.format(css)
    else:
        return css


def oceans16():
    return _get_theme('oceans16.css')


def reset():
    from IPython.display import Javascript, display
    from nglview import js_utils
    display(Javascript("""
    var ele = document.getElementById('nglview_style')
    document.head.removeChild(ele)
    """))


def dark():
    return _get_theme('dark.css')


def light():
    return _get_theme('light.css')



from .utils import widget_utils


class ComponentViewer:
    """Convenient attribute for NGLWidget. See example below.

    Examples
    --------
    >>> view = nv.NGLWidget() # doctest: +SKIP
    ... view.add_trajectory(traj) # traj is a component 0
    ... c = view.add_component(filename) # component 1
    ... c.clear()
    ... c.add_cartoon()
    ... c.add_licorice()
    ... view.remove_component(c)
    """

    def __init__(self, view, index):
        self._view = view
        self._index = index
        widget_utils._add_repr_method_shortcut(self, self._view)
        self._borrow_attribute(self._view, [
            'clear_representations', '_remove_representations_by_name',
            '_update_representations_by_name', 'center_view', 'center',
            'clear', 'set_representations'
        ], ['get_structure_string', 'get_coordinates', 'n_frames'])

    @property
    def id(self):
        return self._view._ngl_component_ids[self._index]

    def set_coordinates(self, coordinates):
        """

        Parameters
        ----------
        coordinates : numpy.ndarray, shape=(3, n_atoms)
        """
        self._view.set_coordinates({self._index: coordinates})

    def hide(self):
        """set invisibility for given components (by their indices)
        """
        self._view._remote_call("setVisibility",
                                target='compList',
                                args=[
                                    False,
                                ],
                                kwargs={'component_index': self._index})
        traj = self._view._get_traj_by_id(self.id)
        if traj is not None:
            traj.shown = False

    def show(self):
        """set invisibility for given components (by their indices)
        """
        self._view._remote_call("setVisibility",
                                target='compList',
                                args=[
                                    True,
                                ],
                                kwargs={'component_index': self._index})

        traj = self._view._get_traj_by_id(self.id)
        if traj is not None:
            traj.shown = True

    def set_position(self, pos):
        """
        Parameters
        ----------
        pos: List-like of float, length = 3
        """
        self._call('setPosition', pos)

    def set_rotation(self, rot):
        """
        Parameters
        ----------
        rot: List-like of float, length = 3
        """
        self._call('setPosition', rot)

    def set_scale(self, scale):
        """
        Parameters
        ----------
        scale: float
        """
        self._call('setScale', scale)

    def add_representation(self, repr_type, selection='all', **kwargs):
        kwargs['component'] = self._index
        self._view.add_representation(repr_type=repr_type,
                                      selection=selection,
                                      **kwargs)

    def _borrow_attribute(self, view, attributes, trajectory_atts=None):
        from functools import partial
        from types import MethodType

        traj = view._get_traj_by_id(self.id)

        for attname in attributes:
            view_att = getattr(view, attname)
            setattr(self, '_' + attname, MethodType(view_att, view))
            self_att = partial(getattr(view, attname), component=self._index)
            setattr(self, attname, self_att)

        if traj is not None and trajectory_atts is not None:
            for attname in trajectory_atts:
                traj_att = getattr(traj, attname)
                setattr(self, attname, traj_att)

    def _call(self, method, *args, **kwargs):
        """

        >>> c = view.add_component('file.pdb') # doctest: +SKIP
        ... c._call('setPosition', [10, 20, 0])
        ... c._call("setRotation", [1, 2, 0])
        """
        kwargs2 = {}
        kwargs2.update(kwargs)
        kwargs2['component_index'] = self._index
        self._view._remote_call(method,
                                target='compList',
                                args=args,
                                kwargs=kwargs2)



import nglview as nv
from nglview.color import ColormakerRegistry as cm
from nglview.color import _ColormakerRegistry
from mock import patch, call, ANY


def test_color_scheme():
    assert _ColormakerRegistry().model_id == _ColormakerRegistry(
    ).model_id  # only one 1 model

    with patch.object(cm, '_call') as mock_call:
        cm.add_scheme('abc', [['blue', '1-10']])
        assert mock_call.call_args_list == [
            call('addSelectionScheme', 'abc', [['blue', '1-10']])
        ]
        mock_call.reset_mock()
        cm.add_scheme(
            'abc', """
              this.colorAtom = function(atom){
              }
        """)
        mock_call.assert_called_once_with("executeCode", ANY)

    # handle before and after display
    cm.add_scheme('abc', [['blue', '1-10']])
    assert cm._msg_ar == cm._msg_q == [{
        'args': ('abc', [['blue', '1-10']]),
        'kwargs': {},
        'methodName': 'addSelectionScheme',
        'type': 'callMethod'
    }]
    cm._ready = True
    assert cm._msg_q == []
    cm._ipython_display_()



from .utils.py_utils import _update_url

SHAPE_EXAMPLES = {
    'mesh':
    """
# add_mesh(position, color, index, normal, name)
>>> shape.add_mesh(
        [ 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 1 ],
        [ 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0 ]
)
""",
    'sphere':
    """
# add_sphere(position, color, radius, name)
>>> shape.add_sphere([0, 0, 9], [1, 0, 0], 1.5)
""",
    'ellipsoid':
    """
# add_ellipsoid(position, color, radius, majorAxis, minorAxis, name)
>>> shape.add_ellipsoid([ 6, 0, 0], [ 1, 0, 0 ], 1.5, [ 3, 0, 0 ], [ 0, 2, 0 ])
""",
    'cylinder':
    """
# add_cylinder(position1, position2, color, radius, name)
>>> shape.add_cylinder( [ 0, 2, 7 ], [ 0, 0, 9 ], [ 1, 1, 0 ], 0.5 )
""",
    'cone':
    """
# add_cone(position1, position2, color, radius, name) 
>>> shape.add_cone( [ 0, 2, 7 ], [ 0, 3, 3 ], [ 1, 1, 0 ], 1.5 )
""",
    'arrow':
    """
# add_arrow(position1, position2, color, radius, name)
>>> shape.add_arrow( [ 0, 2, 7 ], [ 0, 0, 9 ], [ 1, 1, 0 ], 0.5 )
""",
    'label':
    """
# add_text(position, color, size, text)
>>> shape.add_text( [ 10, -2, 4 ], [ 0.2, 0.5, 0.8 ], 0.5, "Hello" )
""",
    'text':
    """
# add_text(position, color, size, text)
>>> shape.add_text( [ 10, -2, 4 ], [ 0.2, 0.5, 0.8 ], 0.5, "Hello" )
""",
    'tetrahedron':
    """
# add_tetrahedron
>>> shape.add_tetrahedron([ 27, 0, 1 ], [ 0, 1, 0 ], 2, [ 0, 1, 1 ], [ 1, 0, 1 ])
""",
    'octahedron':
    """
# add_octahedron
>>> shape.add_octahedron([ 8, 5, 14 ], [ 0, 1, 0 ], 2, [ 0, 1, 1 ], [ 1, 0, 1 ])
""",
    'torus':
    """
# add_torus
>>> shape.add_torus([ 16, 5, 14 ], [ 0, 1, 0 ], 2, [ 0, 1, 1 ], [ 1, 0, 1 ])
"""
}


class Shape:
    """TODO: doc

    Parameters
    ----------
    view : nglview.NGLWidget

    Notes
    -----
    Unstable feature

    Examples
    --------
    >>> import nglview as nv
    >>> view = nv.NGLWidget()
    >>> view
    >>> shape = nv.Shape(view=view)
    >>> # TODO: add example
    >>> shape.add_sphere(...)
    """

    def __init__(self, view):
        self.view = view
        self._make_func(SHAPE_EXAMPLES.keys())

    def _make_func(self, names):
        from types import MethodType

        def make_func(name):
            def func(this, *args):
                args_with_name = [
                    name,
                ] + list(args)
                self.add(*args_with_name)

            func.__doc__ = SHAPE_EXAMPLES[name]
            return func

        for name in names:
            func_name = 'add_' + name
            func = make_func(name)
            setattr(self, func_name, MethodType(func, self))

    @_update_url
    def add(self, *args):
        """

        Examples
        --------
        >>> view.shape.add('text', [0, 4, -1], [0.2, 0.5, 0.8], 2.5, 'Meow')

        See also
        --------
        {ngl_url}
        """

        self.view._add_shape([
            args,
        ])

    def add_buffer(self, name, **kwargs):
        """

        Examples
        --------
        >>> view.shape.add_buffer("cone",
        ...     position1=[0, 0, 0],
        ...     position2=[1, 1, 1],
        ...     color=[1, 0, 0],
        ...     color2=[0, 1, 0],
        ...     radius= [1])
        """
        self.view._remote_call("addBuffer",
                               target="Widget",
                               args=[name],
                               kwargs=kwargs,
                               fire_embed=True)



import os
import os.path
import uuid
from functools import partial
from io import StringIO
from tempfile import NamedTemporaryFile
from urllib.request import urlopen

import numpy as np

from . import config
from .base_adaptor import Structure, Trajectory
from .utils.py_utils import FileManager, tempfolder

__all__ = [
    'FileStructure',
    'TextStructure',
    'RdkitStructure',
    'PdbIdStructure',
    'ASEStructure',
    'BiopythonStructure',
    'IOTBXStructure',
    'IODataStructure',
    'QCelementalStructure',
    'Psi4Structure',
    'OpenbabelStructure',
    'RosettaStructure',
    'ProdyStructure',
    'SimpletrajTrajectory',
    'ProdyTrajectory',
    'MDTrajTrajectory',
    'PyTrajTrajectory',
    'ParmEdTrajectory',
    'MDAnalysisTrajectory',
    'HTMDTrajectory',
    'ASETrajectory',
    'SchrodingerStructure',
    'SchrodingerTrajectory',
    'register_backend',
]


def _get_structure_string(write_method, suffix='.pdb'):
    with NamedTemporaryFile(suffix=suffix) as fh:
        write_method(fh.name)
        return fh.read().decode()


class register_backend:
    def __init__(self, package_name):
        # package_name must match exactly to your Python package
        self.package_name = package_name

    def __call__(self, cls):
        config.BACKENDS[self.package_name] = cls
        return cls


class FileStructure(Structure):
    def __init__(self, path):
        super().__init__()
        self.fm = FileManager(path)
        self.ext = self.fm.ext
        self.params = {}
        self.path = path

    def get_structure_string(self):
        return self.fm.read(force_buffer=True)


class TextStructure(Structure):
    def __init__(self, text, ext='pdb', params={}):
        super().__init__()
        self.path = ''
        self.ext = ext
        self.params = params
        self._text = text

    def get_structure_string(self):
        return self._text


class RdkitStructure(Structure):
    def __init__(self, rdkit_mol, ext="pdb"):
        super().__init__()
        self.path = ''
        self.ext = ext
        self.params = {}
        self._rdkit_mol = rdkit_mol

    def get_structure_string(self):
        from rdkit import Chem
        fh = StringIO(Chem.MolToPDBBlock(self._rdkit_mol))
        return fh.read()


class PdbIdStructure(Structure):
    def __init__(self, pdbid):
        super().__init__()
        self.pdbid = pdbid
        self.ext = "cif"
        self.params = {}

    def get_structure_string(self):
        url = "http://www.rcsb.org/pdb/files/" + self.pdbid + ".cif"
        return urlopen(url).read().decode('utf-8')


class ASEStructure(Structure):
    def __init__(self, ase_atoms, ext='pdb', params={}):
        super().__init__()
        self.path = ''
        self.ext = ext
        self.params = params
        self._ase_atoms = ase_atoms

    def get_structure_string(self):
        return _get_structure_string(self._ase_atoms.write)


class IODataStructure(Structure):
    def __init__(self, obj):
        super().__init__()
        self._obj = obj

    def get_structure_string(self):
        """Require `ase` package
        """
        import ase.io
        with NamedTemporaryFile(suffix='.xyz') as fh:
            self._obj.to_file(fh.name)
            return ASEStructure(ase.io.read(fh.name)).get_structure_string()


class QCelementalStructure(Structure):
    def __init__(self, obj):
        super().__init__()
        self._obj = obj

    def get_structure_string(self):
        """Require `ase` package
        """
        import ase.io
        with NamedTemporaryFile(suffix='.xyz') as fh:
            with open(fh.name, 'w') as fh2:
                fh2.write(self._obj.to_string('xyz'))
            return ASEStructure(ase.io.read(fh.name)).get_structure_string()


class Psi4Structure(QCelementalStructure):
    pass


class OpenbabelStructure(Structure):
    def __init__(self, obj):
        super().__init__()
        self._obj = obj

    def get_structure_string(self):
        import openbabel
        oc = openbabel.OBConversion()
        oc.SetOutFormat('pdb')
        write = partial(oc.WriteFile, self._obj)
        return _get_structure_string(write)


class BiopythonStructure(Structure):
    def __init__(self, entity, ext='pdb', params={}):
        super().__init__()
        self.path = ''
        self.ext = ext
        self.params = params
        self._entity = entity

    def get_structure_string(self):
        from Bio.PDB import PDBIO
        from io import StringIO
        io_pdb = PDBIO()
        io_pdb.set_structure(self._entity)
        io_str = StringIO()
        io_pdb.save(io_str)
        return io_str.getvalue()


class IOTBXStructure(Structure):
    def __init__(self, obj, ext='pdb', params={}):
        """
        obj must have as_pdb_string method
        """
        super().__init__()
        self.path = ''
        self.ext = ext
        self.params = params
        self._mol = obj

    def get_structure_string(self):
        return self._mol.as_pdb_string()


class RosettaStructure(Structure):
    def __init__(self, pose, ext='pdb', params={}):
        # type: (pyrosetta.rosetta.core.pose.Pose, str, Dict) -> None
        super().__init__()
        self.path = ''
        self.ext = ext
        self.params = params
        self._mol = pose

    def get_structure_string(self):
        return _get_structure_string(self._mol.dump_pdb)


@register_backend('prody')
class ProdyStructure(Structure):
    def __init__(self, obj):
        super().__init__()
        self._obj = obj
        self.ext = 'pdb'
        self.params = {}

    def get_structure_string(self):
        import prody

        def write(fname):
            if isinstance(self._obj, prody.Ensemble):
                st = self._obj[0]
            else:
                st = self._obj
            prody.writePDB(fname, st)

        return _get_structure_string(write)


@register_backend('prody')
class ProdyTrajectory(Trajectory, ProdyStructure):
    def __init__(self, obj):
        ProdyStructure.__init__(self, obj)

    @property
    def n_frames(self):
        return self._obj.numConfs()

    def get_coordinates(self, index):
        return self._obj.getConformation(index).getCoords()


@register_backend('simpletraj')
class SimpletrajTrajectory(Trajectory, Structure):
    '''simpletraj adaptor.

    Examples
    --------
    >>> import nglview as nv
    >>> t = nv.SimpletrajTrajectory(nv.datafiles.XTC, nv.datafiles.GRO)
    >>> w = nv.NGLWidget(t)
    >>> w
    '''

    def __init__(self, path, structure_path):
        try:
            import simpletraj
        except ImportError as e:
            raise ImportError(
                "'SimpletrajTrajectory' requires the 'simpletraj' package")
        self.traj_cache = simpletraj.trajectory.TrajectoryCache()
        self.path = path
        self._structure_path = structure_path
        self.ext = os.path.splitext(structure_path)[1][1:]
        self.params = {}
        self.trajectory = None
        self.id = str(uuid.uuid4())
        try:
            self.traj_cache.get(os.path.abspath(self.path))
        except Exception as e:
            raise e

    def get_coordinates(self, index):
        traj = self.traj_cache.get(os.path.abspath(self.path))
        frame = traj.get_frame(index)
        return frame["coords"]

    def get_structure_string(self):
        return open(self._structure_path).read()

    @property
    def n_frames(self):
        traj = self.traj_cache.get(os.path.abspath(self.path))
        return traj.numframes


@register_backend('mdtraj')
class MDTrajTrajectory(Trajectory, Structure):
    '''mdtraj adaptor.

    Example
    -------
    >>> import nglview as nv
    >>> import mdtraj as md
    >>> traj = md.load(nv.datafiles.XTC, nv.datafiles.GRO)
    >>> t = MDTrajTrajectory(traj)
    >>> w = nv.NGLWidget(t)
    >>> w
    '''

    def __init__(self, trajectory):
        self.trajectory = trajectory
        self.ext = "pdb"
        self.params = {}
        self.id = str(uuid.uuid4())

    def get_coordinates(self, index):
        return 10 * self.trajectory.xyz[index]

    @property
    def n_frames(self):
        return self.trajectory.n_frames

    def get_structure_string(self):
        return _get_structure_string(self.trajectory[0].save_pdb)


@register_backend('pytraj')
class PyTrajTrajectory(Trajectory, Structure):
    '''PyTraj adaptor.

    Example
    -------
    >>> import nglview as nv
    >>> import pytraj as pt
    >>> traj = pt.load(nv.datafiles.TRR, nv.datafiles.PDB)
    >>> t = nv.PyTrajTrajectory(traj)
    >>> w = nv.NGLWidget(t)
    >>> w
    '''

    def __init__(self, trajectory):
        self.trajectory = trajectory
        self.ext = "pdb"
        self.params = {}
        self.id = str(uuid.uuid4())

    def get_coordinates(self, index):
        return self.trajectory[index].xyz

    @property
    def n_frames(self):
        return self.trajectory.n_frames

    def get_structure_string(self, index=0):
        return _get_structure_string(self.trajectory[index:index + 1].save)


@register_backend('parmed')
class ParmEdStructure(Structure):
    def __init__(self, structure):
        self._structure = structure
        self.only_save_1st_model = True

    def get_structure_string(self):
        # only write 1st model
        with NamedTemporaryFile(suffix='.pdb') as fh:
            if self.only_save_1st_model:
                self._structure.write_pdb(
                    fh.name, coordinates=self._structure.coordinates)
            else:
                self._structure.write_pdb(fh.name)
            return fh.read().decode()


@register_backend('parmed')
class ParmEdTrajectory(Trajectory, ParmEdStructure):
    '''ParmEd adaptor.
    '''

    def __init__(self, trajectory):
        ParmEdStructure.__init__(self, trajectory)
        self.trajectory = self._structure = trajectory
        self.ext = "pdb"
        self.params = {}
        # only call get_coordinates once
        self._xyz = trajectory.get_coordinates()
        self.id = str(uuid.uuid4())

    def get_coordinates(self, index):
        return self._xyz[index]

    @property
    def n_frames(self):
        return len(self._xyz)


@register_backend('MDAnalysis')
class MDAnalysisTrajectory(Trajectory, Structure):
    '''MDAnalysis adaptor.

    Can take either a Universe or AtomGroup.

    Example
    -------
    >>> import nglview as nv
    >>> import MDAnalysis as mda
    >>> u = mda.Universe(nv.datafiles.GRO, nv.datafiles.XTC)
    >>> prot = u.select_atoms('protein')
    >>> t = nv.MDAnalysisTrajectory(prot)
    >>> w = nv.NGLWidget(t)
    >>> w
    '''

    def __init__(self, atomgroup):
        self.atomgroup = atomgroup
        self.ext = "pdb"
        self.params = {}
        self.id = str(uuid.uuid4())

    def get_coordinates(self, index):
        self.atomgroup.universe.trajectory[index]
        xyz = self.atomgroup.atoms.positions
        return xyz

    @property
    def n_frames(self):
        return self.atomgroup.universe.trajectory.n_frames

    def get_structure_string(self):
        try:
            import MDAnalysis as mda
        except ImportError:
            raise ImportError(
                "'MDAnalysisTrajectory' requires the 'MDAnalysis' package")
        u = self.atomgroup.universe
        u.trajectory[0]
        f = mda.lib.util.NamedStream(StringIO(), 'tmp.pdb')
        atoms = self.atomgroup.atoms
        # add PDB output to the named stream
        with mda.Writer(f, atoms.n_atoms, multiframe=False) as W:
            W.write(atoms)
        # extract from the stream
        pdb_string = f.read()
        return pdb_string


@register_backend('htmd')
class HTMDTrajectory(Trajectory):
    '''HTMD adaptor.

    Takes a Molecule object.

    Example
    -------
    >>> import nglview as nv
    >>> from htmd import Molecule
    >>> mol = Molecule(nv.datafiles.PDB)
    >>> t = nv.HTMDTrajectory(mol)
    >>> w = nv.NGLWidget(t)
    >>> w
    '''

    def __init__(self, mol):
        self.mol = mol
        self.ext = "pdb"
        self.params = {}
        self.id = str(uuid.uuid4())

    def get_coordinates(self, index):
        return np.squeeze(self.mol.coords[:, :, index])

    @property
    def n_frames(self):
        return self.mol.numFrames

    def get_structure_string(self):
        return _get_structure_string(self.mol.write)


@register_backend('ase')
class ASETrajectory(Trajectory, Structure):
    '''asetraj adaptor.

    Examples
    --------
    >>> import nglview as nv
    >>> from ase.io.trajectory import Trajectory
    >>> traj = Trajectory(nv.datafiles.ASE_Traj)
    >>> t = nv.ASETrajectory(traj)
    >>> w = nv.NGLWidget(t)
    >>> w.add_spacefill()
    >>> w
    '''

    def __init__(self, ase_traj):
        self.ext = 'pdb'
        self.params = {}
        self.trajectory = ase_traj
        self.id = str(uuid.uuid4())

    def get_coordinates(self, index):
        return self.trajectory[index].positions

    def get_structure_string(self):
        return _get_structure_string(self.trajectory[0].write)

    @property
    def n_frames(self):
        return len(self.trajectory)


@register_backend('schrodinger')
class SchrodingerStructure(Structure):
    '''Only read first structure
    '''

    def __init__(self, structure, ext="pdb"):
        super().__init__()
        self.path = ''
        self.ext = ext
        self.params = {}
        self._schrodinger_structure = structure

    def get_structure_string(self):
        # NOTE: For some reasons, _get_structure_string always return
        # empty string in this case.
        with tempfolder():
            pdb_fn = 'tmp.pdb'
            self._schrodinger_structure.write(pdb_fn)
            with open(pdb_fn) as fh:
                content = fh.read()
        return content

        return _get_structure_string(self._schrodinger_structure.write)


@register_backend('schrodinger')
class SchrodingerTrajectory(SchrodingerStructure, Trajectory):
    """Require `parmed` package.
    """

    def __init__(self, structure, traj):
        super().__init__(structure)
        self._traj = traj

    @property
    def n_frames(self):
        return len(self._traj) if self._traj else 1

    def get_coordinates(self, index):
        return self._traj[index].pos()

    def get_structure_string(self):
        """Require `parmed` package.
        """
        import parmed as pmd
        c = self._schrodinger_structure
        s = pmd.Structure()
        fsys = c.fsys_ct if hasattr(c, 'fsys_ct') else c
        for atom in fsys.atom:
            parm_atom = pmd.Atom(name=atom.pdbname.strip(),
                                 atomic_number=atom.atomic_number)
            s.add_atom(parm_atom,
                       atom.pdbres.strip(),
                       atom.resnum,
                       chain=atom.chain)
        s.coordinates = fsys.getXYZ()
        return ParmEdStructure(s).get_structure_string()

    @classmethod
    def from_files(cls, cms_fname, traj_fname):
        from schrodinger.application.desmond.packages import topo, traj
        _, model = topo.read_cms(cms_fname)
        traj = traj.read_traj(traj_fname)
        return cls(model, traj)



from ipywidgets import DOMWidget
from traitlets import Unicode, Bool, observe
from .base import BaseWidget
from ._frontend import __frontend_version__
from IPython.display import display
import time


COLOR_SCHEMES = [
    " ", "picking", "random", "uniform", "atomindex", "residueindex",
    "chainindex", "modelindex", "sstruc", "element", "resname", "bfactor",
    "hydrophobicity", "value", "volume", "occupancy"
]

_USER_COLOR_DICT = {}



class _ColorScheme:
    _color_dict = {}

    def __init__(self, args, label):
        # FIXME: validate `args`
        self._color_scheme = args
        self._label = f'user_{label}'
        _USER_COLOR_DICT[self._label] = self._color_scheme

    @property
    def data(self):
        return {'data': self._color_scheme, 'label': self._label}


def _singleton(cls):
    # https://www.python.org/dev/peps/pep-0318/#examples
    instances = {}
    def getinstance():
        if cls not in instances:
            instances[cls] = cls()
        return instances[cls]
    return getinstance


@_singleton
class _ColormakerRegistry(BaseWidget):
    _view_name = Unicode("ColormakerRegistryView").tag(sync=True)
    _view_module = Unicode("nglview-js-widgets").tag(sync=True)
    _view_module_version = Unicode(__frontend_version__).tag(sync=True)
    _model_name = Unicode("ColormakerRegistryModel").tag(sync=True)
    _model_module = Unicode("nglview-js-widgets").tag(sync=True)
    _model_module_version = Unicode(__frontend_version__).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            get_ipython() # only display in notebook
            display(self)
        except NameError:
            pass

    @observe("_ready")
    def _on_ready(self, change):
        if change.new:
            while self._msg_q:
                msg = self._msg_q.pop(0)
                self.send(msg)

    def _ipython_display_(self, **kwargs):
        if self._ready:
            return str(self)
        super()._ipython_display_(**kwargs)

    def add_selection_scheme(self, scheme_id, arg):
        """
        Examples
        --------
        >>> ColormakerRegistry.add_selection_scheme("my_custom_scheme",
                [['blue', '1-10']])
        >>> view.add_cartoon(color="my_custom_scheme")
        """
        self._call("addSelectionScheme", scheme_id, arg)

    def add_scheme_func(self, scheme_id, func_str):
        """

        Examples
        --------
        >>> func_str = '''
             this.atomColor = function (atom) {
             if (atom.serial < 1000) {
               return 0x0000FF // blue
             } else if (atom.serial > 2000) {
               return 0xFF0000 // red
             } else {
               return 0x00FF00 // green
             }
             }
         '''
        >>> ColormakerRegistry.add_scheme_func('awesome', func_str)
        >>> view.add_cartoon(color='awesome')
        """

        code = """
            var schemeId = NGL.ColormakerRegistry.addScheme(function (params) {
                %s
            })
            
            this._updateId(schemeId, '%s')
        """ % (func_str, scheme_id)
        self._js(code)

    def add_scheme(self, scheme_id, obj):
        """
        Parameters
        ----------
        obj: List of List or str (of JS function)
        """
        if isinstance(obj, list):
            self.add_selection_scheme(scheme_id, obj)
        elif isinstance(obj, str):
            self.add_scheme_func(scheme_id, obj)
        else:
            raise ValueError(f"{obj} must be either list of list or string")


ColormakerRegistry = _ColormakerRegistry()



import base64
import bz2
import gzip
import os
import tempfile
from contextlib import contextmanager
from shutil import rmtree
from zipfile import ZipFile

from ..data_source import DatasourceRegistry

__all__ = [
    'encode_base64', 'decode_base64', 'seq_to_string', '_camelize',
    '_camelize_dict', 'get_colors_from_b64', 'display_gif'
]


def get_positive_index(index, size):
    """Used for negative indexing"""
    if index < 0:
        index = size + index
        if index < 0:
            raise ValueError("index is out of range")
    if index >= size:
        raise ValueError("index is out of range")
    return index


def click(button):
    button._handle_button_msg(None, dict(event='click'), None)


def submit(widget_text):
    widget_text._handle_string_msg(None, dict(event='submit'), None)


def _update_url(func):
    from nglview.default import NGL_BASE_URL
    func.__doc__ = func.__doc__.format(ngl_url=NGL_BASE_URL)
    return func


def encode_base64(arr, dtype='f4'):
    arr = arr.astype(dtype)
    return base64.b64encode(arr.data).decode('utf8')


def decode_base64(data, shape, dtype='f4'):
    import numpy as np
    decoded_str = base64.b64decode(data)
    return np.frombuffer(decoded_str, dtype=dtype).reshape(shape)


def get_name(obj, kwargs):
    name = kwargs.pop('name', str(obj))
    if name.startswith('<nglview.'):
        name = name.split()[0].strip('<')
    return name


def display_gif(fn):
    from IPython import display
    return display.HTML(f'<img src="{fn}">')


@contextmanager
def tempfolder():
    """run everything in temp folder
  """
    my_temp = tempfile.mkdtemp()
    cwd = os.getcwd()
    os.chdir(my_temp)
    yield
    os.chdir(cwd)
    rmtree(my_temp)


def get_repr_names_from_dict(repr_dict, component):
    """
    
    Parameters
    ----------
    repr_dict : dict
        see view._repr_dict
    component : int
        component index
    """

    try:
        this_repr_dict = repr_dict[str(component)]
        return [
            this_repr_dict[str(key)]['type']
            for key in sorted(this_repr_dict.keys())
        ]
    except KeyError:
        return []


def get_colors_from_b64(b64_image):
    """

    Examples
    --------
    >>> view.render_image()
    >>> get_colors_from_b64(view._image_data)

    Returns
    -------
    list of tuple 
    """
    # should install PIL
    # py3

    from PIL import Image
    import io, base64

    fp = io.BytesIO(base64.b64decode(b64_image))
    image = Image.open(fp)

    return image.getcolors(int(1E6))


def seq_to_string(seq):
    """e.g. convert [1, 3, 5] to "@1,3,5"
    """
    if isinstance(seq, str):
        return seq
    else:
        # assume 1D array
        return "@" + ",".join(str(s) for s in seq)


def _camelize(snake):
    """
    
    Examples
    --------
    >>> _camelize('remote_call')
    remoteCall
    >>> _camelize('remoteCall')
    remoteCall
    """
    words = snake.split('_')
    return words[0] + "".join(x.title() for x in words[1:])


def _camelize_dict(kwargs):
    return {_camelize(k): v for k, v in kwargs.items()}


def snakify(from_camel):
    """ 
    
    Examples
    --------
    >>> snakify("heyThere") -> "hey_there"
    """
    import re
    return re.sub('([A-Z]{1})', r'_\1', from_camel).lower()


class FileManager:
    """FileManager is for internal use.

    If file is in the current folder or subfoler, use filename
    If not, open content

    Parameters
    ----------
    src : str or file-like object
        filename
    compressed : None or bool, default None
        user can specify if the given file is compressed or not.
        if None, FileManager will detect based on file extension
    """

    def __init__(self, src, compressed=None, ext=None):
        self.src = src
        self.cwd = os.getcwd()
        self._compressed = compressed
        self._ext = ext
        self.unzip_backend = dict(gz=gzip, bz2=bz2, zip=ZipFile)

    def read(self, force_buffer=False):
        """prepare content to send to NGL
        """
        if self.use_filename and not force_buffer:
            return os.path.relpath(self.src)
        else:
            if self.compressed_ext:
                return self.unzip_backend[self.compressed_ext].open(
                    self.src).read()
            elif hasattr(self.src, 'read'):
                return self.src.read()
            else:
                if self.is_filename:
                    mode = 'rb' if self.is_binary else 'r'
                    with open(self.src, mode) as fh:
                        return fh.read()
                else:
                    return self.src

    @property
    def is_compressed(self):
        '''naive detection
        '''
        if self._compressed is None:
            if self.is_filename or self.is_url:
                return (self.src.endswith('gz') or self.src.endswith('zip')
                        or self.src.endswith('bz2'))
            else:
                return False
        else:
            return self._compressed

    @property
    def compressed_ext(self):
        if self.is_compressed and self.is_filename:
            return self.src.split('.')[-1]
        else:
            return ''

    @property
    def use_filename(self):
        if hasattr(self.src, 'read'):
            return False
        else:
            if self.is_filename:
                cwd = os.path.realpath(os.getcwd())
                root_path = os.path.realpath(
                    os.path.dirname(os.path.abspath(self.src)))
                return (cwd in root_path)
            return False

    @property
    def ext(self):
        if self._ext is not None:
            return self._ext
        else:
            if hasattr(self.src, 'read') or (not self.is_filename
                                             and not self.is_url):
                raise ValueError(
                    "you must provide file extension if using file-like object or text content"
                )
            if self.is_compressed:
                return self.src.split('.')[-2]
            else:
                return self.src.split('.')[-1]

    @property
    def is_filename(self):
        if hasattr(self.src, 'read'):
            return False
        else:
            return os.path.isfile(self.src)

    @property
    def is_binary(self):
        binary_exts = ["mmtf", "dcd", "mrc", "ccp4", "map", "dxbin"]
        return self.ext.lower() in binary_exts

    @property
    def is_url(self):
        url_ext = ['http', 'rcsb://', 'data://'] + \
                  [f"{k}://" for k in DatasourceRegistry.sources]
        return (isinstance(self.src, str)
                and self.src.startswith(tuple(url_ext)))
